//
//	Includes
//
#include	"MCF52233.h"	// Common headder for ColdFire
#include	"ProtoType.h"	// Prototype headder for SilentMoon system call

//
//	Prototypes
//
UINT8 SenseByte(void);	// Sense char of UART
UINT8 GetByte(void);		// Get char from UART
void	PutByte(UINT8);		// Put char for UART
void	InitDevice(UINT8);	// Device init
void	SerialMain(UINT32);	// Main loop
void	SetupData(void);		// Setup UDP packet
void	DataIndication(void);	// Handle incoming UDP packet
__declspec(interrupt) void TimerHandler(void);	// Timer handler
__declspec(interrupt) void SciHandler(void);		// UART handler

//
//	Constants (change by user)
//
#define	SPEED			32		// Set default speed 57600
//115200=16,57600=32,38400=49,28800=65,19200=97,14400=130,9600=195
#define	LOOPTIME	10		// Default main loop time (10ms)
#define	PORTNUM		9999	// Default port number

//
//	Buffer size constant 
//
#define	ETHERBUF	1024	// UDP payload size
#define	RX_SIZE		1024	// UART receive buffer
#define	RX_MASK		0x3ff	// Mask for 1024 byte ring buffer
#define	TX_SIZE		1024	// UART transmit buffer
#define	TX_MASK		0x3ff	// Mask for 1024 byte ring buffer

//
//	UDP Packet data struct
//
typedef	struct 
{
	UINT16	Size;		// Payload size
	UINT8	AckSeq;		// Acknowledged seq number
	UINT8	SeqNum;		// Packet seq number

	UINT8	data[ETHERBUF];	// Data buffer
}	PacketBuffer_t;	// Typedef name of UDP packet buffer

//
//	Vars
//
PacketBuffer_t	*TxPacketBuffer;	// Transmit ether packet buffer
PacketBuffer_t	*RxPacketBuffer;	// Recieve ether packet buffer
UINT8	*RxBuf;			// Receive ring buffer (UART)
UINT8	*TxBuf;			// Transmit ring buffer (UART)
UINT16	RxRingIn;		// Receive ring in pointer (UART)
UINT16	RxRingOut;	// Receive ring out pointer (UART)
UINT16	TxRingIn;		// Transmit ring in pointer (UART)
UINT16	TxRingOut;	// Transmit ring out pointer (UART)
UINT8	TxSeqNum;			// My packet seq number
UINT8	RxSeqNum;			// Received packet seq number
UINT8	RxLastNum;		// Last processed packet seq number
INT8 Soc;						// Socket handle for UDP send/receive
volatile UINT8 TickCount;	// Timer tick of 1ms timer

//
//	Timer handler (1ms tick)
//
__declspec(interrupt) void TimerHandler(void)
{
	MCF_DTIM0_DTER = 2;	// Clear timer interrupt flag
	TickCount++;				// Advance ticke work
}

//
//	UART interrupt handler
//
__declspec(interrupt) void SciHandler(void)
{
UINT8 st, ch;		// Work for status & data
	st = 	MCF_UART0_UISR;	// Get UART interrupt status register
	if(st & MCF_UART_UISR_FFULL_RXRDY)	// Receive data?
	{		// Yes, Serial data has arrived
		ch = MCF_UART0_URB;	// Get arrived data from data buffer
		if(RxRingOut != (RxRingIn + 1) & RX_MASK)	// Ring buffer has space to put?
		{
			RxBuf[RxRingIn] =	ch;	// Store arrived data to ring buffer
			RxRingIn = (UINT16)((RxRingIn + 1) & RX_MASK);// Advance ring in pointer
			if(((RxRingIn - RxRingOut) & RX_MASK) > RX_MASK / 2)	// Half of ring buffer occupied?
				MCF_UART0_UOP0 = 1;	// Set RTS off for flow control(disable flow)
		}
		return;	// Don't go to transmit handling(another intterrupt happen when needed)
	}
	if(st & MCF_UART_UISR_TXRDY)	// Trnsmitter is ready to send?
	{	// Yes,then send one char and stop when all data is sent
		if(TxRingIn == TxRingOut)	// Transmit ring buffer is empty?
		{	// Yes,no more data to transmit then disable Tx interrupt
			MCF_UART0_UIMR = MCF_UART_UIMR_FFULL_RXRDY;// Clear Tx interrupt bit
			return;	// No more data to send then stop Tx intterrupt and return
		}
		MCF_UART0_UTB = TxBuf[TxRingOut];	// Send one char from transmit ring buffer
		TxRingOut = (UINT16)(((TxRingOut + 1) & TX_MASK));	// Advance ring out index
	}
}

//
//	Sense data of UART
//
UINT8 SenseByte(void)
{
	return (UINT8)(RxRingIn != RxRingOut);	// Check ring buffer has data
}

//
//	Get one char from ring buffer
//
UINT8 GetByte(void)
{
UINT8 getc;	// char save work
	while(!SenseByte())SystemSleep();	// Wait until UART data has come
  getc = RxBuf[RxRingOut];	// Get one char from ring buffer
	RxRingOut = (UINT16)((RxRingOut + 1) & RX_MASK);// Advance ring buffer out pointer
	if(((RxRingIn - RxRingOut) & RX_MASK) < RX_SIZE / 4)// Data amout is less than 1/4?
		MCF_UART0_UOP1 = 1;		// set RTS on for flow control (flow enable)
	return getc;	// Return with UART input data
}

//
//	Put one char to send queue
//
void PutByte(UINT8 ch)
{
	if(TxRingOut != ((TxRingIn + 1) & TX_MASK))	// Ring buffer has space to put?
	{
		TxBuf[TxRingIn] = ch;	// Put data to transmit ring buffer
		TxRingIn = (UINT16)((TxRingIn + 1) & TX_MASK);	// Advance ring in pointer
	}
}

//
//	Init timer & UART
//
void InitDevice(UINT8 baud)
{
//	Setup interval timer
	*(UINT32*)0x2000014c = (UINT32)TimerHandler;	// Set timer handler address(interrupt 19)
	MCF_INTC0_ICR19 = MCF_INTC_ICR_IL(5) | MCF_INTC_ICR_IP(5);// Set interrupt priority (5)
	MCF_INTC0_IMRL &= ~MCF_INTC_IMRL_INT_MASK19;	// Clear interrupt 19 mask bit(allow int)
	MCF_DTIM0_DTMR = 0x1a;		// Use internal clock & repeat & enable interrupt
	MCF_DTIM0_DTRR = 60000 - 1;	// Set timer interval to 1ms
	MCF_DTIM0_DTMR |= 1;			// Enable timer module
//	Allocate UART ring buffer for Tx & Rx
	RxBuf = MemoryAlloc(RX_SIZE);	// Allocate Rx ring buffer
	TxBuf = MemoryAlloc(TX_SIZE);	// Allocate Tx ring buffer
	RxRingIn = 0;	// Clear ring in pointer for receive
	RxRingOut = 0;// Clear ring out pointer for receive
	TxRingIn = 0;	// Clear ring in pointer for transmit
	TxRingOut = 0;// Clear ring out pointer for transmit
//	Setup UART registers
	MCF_GPIO_PUAPAR = 0		// Port purpose setting
		| MCF_GPIO_PUAPAR_URXD0_URXD0	// Use Rx pin for UART (not for GPIO)
		| MCF_GPIO_PUAPAR_UTXD0_UTXD0	// Use Tx pin for UART
		| MCF_GPIO_PUAPAR_URTS0_URTS0	// Use RTS pin for UART
		| MCF_GPIO_PUAPAR_UCTS0_UCTS0;// Use CTS pin for UART
	MCF_UART0_UCR = MCF_UART_UCR_RESET_TX;// Reset Transmitter
	MCF_UART0_UCR = MCF_UART_UCR_RESET_RX;// Reset Receiver
	MCF_UART0_UCR = MCF_UART_UCR_RESET_MR;// Reset Mode register write sequence
	MCF_UART0_UMR1 = 0				//	Set serial data format
		| MCF_UART_UMR_PM_NONE	// No parity
		| MCF_UART_UMR_BC_8;		// 8 bit data
	MCF_UART0_UMR2 = 0				//	Access UMR2 after UMR1 write(sequencial)
		| MCF_UART_UMR_SB_STOP_BITS_1;	// 1 stop bit
	MCF_UART0_UCSR = 0				// Set Rx and Tx uses SYSTEM CLOCK
		| MCF_UART_UCSR_RCS_SYS_CLK		// Rx uses system clock
		| MCF_UART_UCSR_TCS_SYS_CLK;	// Tx uses system clock
	MCF_UART0_UIMR = 0				// Allow Rx arrive interrupt
		| MCF_UART_UIMR_FFULL_RXRDY;
	MCF_UART0_UOP1 = 1;				// Set RTS ON to enable flow control
	MCF_UART0_UBG1 = 0;				// Set baud rate register 1 (high byte)
	MCF_UART0_UBG2 = baud;		// Set baud rate register 2 (low byte)
	*(UINT32*)0x20000134 = (UINT32)SciHandler;// Set UART handler address(interrupt 17)
	MCF_INTC0_ICR13 = MCF_INTC_ICR_IL(2);	// Set interrupt priority (2)
	MCF_INTC0_IMRL &= ~MCF_INTC_IMRL_INT_MASK13;	// Clear interrupt 17 mask bit (enable)
	MCF_UART0_UCR = 0					// Finally enable Rx and Tx
		| MCF_UART_UCR_TX_ENABLED		// Set Tx enable bit
		| MCF_UART_UCR_RX_ENABLED;	// Set Rx enable bit
}

//
//	Setup UDP packet
//
void SetupData(void)
{
UINT16	index;
	TxPacketBuffer->AckSeq = RxSeqNum;	// Update received seq num for acknowledge
	if(TxPacketBuffer->Size)		// Last UDP data confirmed?
		return;					// No then re-send same UDP packet (error retry)
	for(index = 0;;)	// Data setup loop
	{
		if(SenseByte() == 0)	// Check UART data has come?
			break;				// No more data then exit setup loop
		TxPacketBuffer->data[index]	= GetByte();	// Store char to packet buffer
		TxPacketBuffer->Size = ++index;	// Advance index and payload size
		if(index == ETHERBUF)	// Pay load size limit?
			break;	// Exit data setup loop when payload is full
	}
	TxPacketBuffer->Size = index;	// Set payload length
	if(++TxSeqNum == 0)	// Increment Packet seq number for new UDP packet
		TxSeqNum++;		// If it's 0 then force to 1
	TxPacketBuffer->SeqNum = TxSeqNum;	// Set Send Seq num
}

//
//	Handle incoming UDP packet
//
void DataIndication(void)
{
UINT16	i;	// Loop counter work
	if(RxPacketBuffer->AckSeq == TxSeqNum)	// Ack seq num matches sent seq num?
		TxPacketBuffer->Size = 0;	// Clear size for next packet send(acknowledged)
	if((MCF_UART0_UIP & 1)	// Check CTS off for flow control
	|| ((TxRingIn - TxRingOut) & TX_MASK) + RxPacketBuffer->Size >= TX_SIZE)
// Check recieved payload will fit Tx ring buffer?(too big payload?)
	{	// Noway to put received data to UART transmit ring buffer
		RxSeqNum = 0; // Force request resend UDP buffer(CTS off or Tx ring is full)
		return;		// Return (not set data to Tx ring buffer)
	}
	RxSeqNum = RxPacketBuffer->SeqNum; // Save recieved packet seq number
	if(RxSeqNum == RxLastNum)	// Already processed last time?
		return;			// Return if it was duplicate packet (dump packet)
	RxLastNum = RxSeqNum;	// Save processed packet seq num
	if(RxPacketBuffer->Size)	// Received packet has payload?
	{	// Yes then put UDP payload to Tx ring buffer for UART transmittion
		for(i = 0; i < RxPacketBuffer->Size; i++)	// Send payload data to Tx ring
			PutByte(RxPacketBuffer->data[i]);	// Put received char to Tx ring buffer
		MCF_UART0_UIMR = MCF_UART_UIMR_FFULL_RXRDY | MCF_UART_UIMR_TXRDY;
// Enable UART tx interrupt and Tx ready interrupt will happen		
	}
}

//
//	Network serial cable main
//
void SerialMain(UINT32 ip)
{
//	Setup parameters
	InitDevice(SPEED);	// Init timer & UART and set baud rate
	TxPacketBuffer = MemoryAlloc(sizeof(PacketBuffer_t));	// Alloc UDP send buffer
	Soc = CreateSocket(0);	// Create UDP socket
	Bind(Soc, PORTNUM, 0);	// Set source port number
// Start main loop
	for(;;)	// Repeat forever
	{	// Transmit UDP and wait UDP reply 
		TickCount = 0;	// Reset time
		SetupData();		// Setup UDP packet
		SendTo(Soc, ip, PORTNUM,	// Send UDP packet 
			(UINT8*)TxPacketBuffer, (UINT16)(TxPacketBuffer->Size + 4));// Pointer & length
		while(TickCount < LOOPTIME - 1)	// Wait reply packet until 1ms before main loop time
		{
			if(RecvFrom(Soc, 0) >= 0)	// Check UDP packet arrival
			{	// Yes,UDP data has arrived
				RxPacketBuffer = (PacketBuffer_t*)GetReceiveBuffer(Soc, 1);// Get UDP data buffer
				DataIndication();	// Handle incoming UDP packet
				MemoryFree(RxPacketBuffer);	// Release recieved UDP data buffer
				break;	// Exit while loop
			}
			SystemSleep();	// Allow another thread to execute their task
		}
		while(TickCount < LOOPTIME)	// Wait until main loop time
			SystemSleep();	// Allow another thread to execute their task
	}
}
